!=======================================================================
!                     MOLECULAR/ATOMIC LINE MODULE
!
! NOTE:
!   The molecular line data file format is that of the Leiden Atomic and
!   Molecular Database (LAMBDA), see
!   http://www.strw.leidenuniv.nl/%7Emoldata/.  The paper belonging to that
!   data is: Schöier, F.L., van der Tak, F.F.S., van Dishoeck E.F., Black,
!   J.H. 2005, A&A 432, 369-379.
!
! NOTE:
!   Since not all lines or levels of an atomic/molecular data file are
!   necessarily used, each level has a 'system index' (meaning the 
!   index for access of the arrays in this module) and a 'true index'
!   (by which the level can be found in the atomic/molecular data file).
!   Also the lines have 'system index' and 'true index'.
!
!=======================================================================
module lines_module
use amr_module
use rtglobal_module
use ioput_module
use mathroutines_module
use constants_module
!
! Warning that due to insufficient spatial resolution of the model
! there are locations where the Doppler shift between two cells is
! larger than the local line width, which means that a line could
! accidently be 'lost' in the ray-tracing. 
!
logical :: lines_warn_lineleap
!
! The maximum nr of levels and lines, used for the array sizes
!
integer :: lines_maxnrlevels=0
integer :: lines_maxnrlines=0
!
! The number of different atomic/molecular species included
!
integer :: lines_nr_species=0
!
! Some parameters that the main program can use for specifying the
! frequency array. If all are 0, then the frequencies array is read
! from a file.
!
integer :: lines_user_nrfreq=0
double precision :: lines_user_widthkms=0.d0
double precision :: lines_user_kms0=0.d0
integer :: lines_user_ispec=0
integer :: lines_user_iline=0
!
! Some defaults for these user defined things
! 
double precision :: lines_user_widthkms_default = 10.d0
integer :: lines_user_nrfreq_default = 40
!
! For each species, the species name. This is also used for finding
! the file name of that species, so please use names that can be used
! as parts of file names. Also use no spaces. 
!
character*80, allocatable :: lines_speciesname(:)
!
! For each species, the style of the input file:
!  leiden  = The input format of the LAMBDA database:
!            http://www.strw.leidenuniv.nl/~moldata/
!
character*80, allocatable :: lines_filestyle(:)
!
! For each species, the available number of levels and lines.
!
integer, allocatable :: lines_nrlevels(:)
integer, allocatable :: lines_nrlines(:)
!
! Same as above, but now only the number that are actually used
! (a subset of the total)
!
integer, allocatable :: lines_nrlevels_subset(:)
integer, allocatable :: lines_nrlines_subset(:)
integer :: lines_nrlevels_subset_max=0
integer :: lines_nrlines_subset_max=0
!
! The list of levels and lines actually used 
!
integer, allocatable :: lines_levels_subset(:,:)
integer, allocatable :: lines_lines_subset(:,:)
!
! For each species, the molecular/atomic weight in units of
! the proton mass
!
double precision, allocatable :: lines_umass(:)
double precision :: lines_umass_min=0.d0
double precision :: lines_umass_max=0.d0
!
! For each level of each species, the energy
!
double precision, allocatable :: lines_level_energy_cminv(:,:)
!
! For each level of each species, the degeneracy parameter
!
double precision, allocatable :: lines_level_gdeg(:,:)
!
! For each level of each species, the name of the level
!
character*256, allocatable :: lines_level_name(:,:)
!
! For each line of each species, the A_{up,down}, B_{up,down} 
! and B_{down,up} Einstein coefficients.
!
double precision, allocatable :: lines_aud(:,:)
!
! For each line of each species, the frequency of the line in Hertz
!
double precision, allocatable :: lines_nu0(:,:)
!
! Each line connects an upper and a lower level. 
!
integer, allocatable :: lines_levelup(:,:)
integer, allocatable :: lines_leveldown(:,:)
!
! Same as above, but now the index of the subset of levels
! This is only used if we have level populations stored
! at each position in space, and for reasons of memory
! saving we only store the subset.
!
integer, allocatable :: lines_levelup_subsetindex(:,:)
integer, allocatable :: lines_leveldown_subsetindex(:,:)
!
! Now the big arrays with the level populations, only for the subset of
! active levels.
!
double precision, allocatable :: lines_levelpop(:,:,:)
!
! (For LTE, LVG, non-LTE internal computation): the number density of the
! various chemical species. This is not necessary if you read the level
! populations directly using read_levelpop().
!
double precision, allocatable :: gas_chemspec_numberdens(:,:)
!
! (For LTE, LVG, non-LTE internal computation): The partition function
! as a function of temperature
!
integer :: lines_partition_ntempint=1000
integer :: lines_partition_ntempmax=1000
integer, allocatable :: lines_partition_ntemp(:)
double precision :: lines_partition_temp0=0.1d0
double precision :: lines_partition_temp1=1d5
double precision, allocatable :: lines_partition_temperature(:,:)
double precision, allocatable :: lines_partition_function(:,:)
!
! Another big array for the velocity field in cm/s
! NOTE: This is not particularly associated to lines, but since it is 
!       only relevant for the lines we put this array here in the line module
!
double precision, allocatable :: gasvelocity(:,:)
!
! Another big array for the microturbulence
!
double precision, allocatable :: lines_microturb(:)
!
! Margin factor for line width, for determining which lines contribute
! to any given frequency
!
double precision :: lines_widthmargin = 12.d0
!
! Maximum fraction of a line width the doppler shift between two neighboring
! cells may be before an alarm is set off.
! !REMOVED FOR NOW!
!
! double precision :: lines_maxdoppler = 0.3d0 
!
! A set of arrays for the ray tracing 
!
double precision, allocatable :: lines_ray_levpop(:,:,:)
double precision, allocatable :: lines_ray_temp(:)
double precision, allocatable :: lines_ray_turb(:)
double precision, allocatable :: lines_ray_doppler(:)
!
! For the line src and alpha
!
integer, allocatable :: active_lines(:,:)
integer, allocatable :: active_nrlines(:)
integer, allocatable :: active_levels(:,:)
integer, allocatable :: active_levels_subsetindex(:,:)
integer, allocatable :: active_nrlevels(:)
integer, allocatable :: active_dummy(:)
!
! Variables for maximum doppler shift
!
double precision :: lines_maxveloc=0.d0,lines_maxturbc=0.d0
double precision :: lines_maxtempc=0.d0,lines_maxshift=0.d0
!
contains


!-------------------------------------------------------------------
!                 CLEANUP THE LINE MODULE DATA
!-------------------------------------------------------------------
subroutine lines_cleanup()
  if(allocated(lines_nrlevels)) deallocate(lines_nrlevels)
  if(allocated(lines_nrlines)) deallocate(lines_nrlines)
  if(allocated(lines_nrlevels_subset)) deallocate(lines_nrlevels_subset)
  if(allocated(lines_nrlines_subset)) deallocate(lines_nrlines_subset)
  if(allocated(lines_levels_subset)) deallocate(lines_levels_subset)
  if(allocated(lines_lines_subset)) deallocate(lines_lines_subset)
  if(allocated(lines_umass)) deallocate(lines_umass)
  if(allocated(lines_speciesname)) deallocate(lines_speciesname)
  if(allocated(lines_filestyle)) deallocate(lines_filestyle)
  if(allocated(lines_nu0)) deallocate(lines_nu0)
  if(allocated(lines_level_gdeg)) deallocate(lines_level_gdeg)
  if(allocated(lines_level_name)) deallocate(lines_level_name)
  if(allocated(lines_level_energy_cminv)) deallocate(lines_level_energy_cminv)
  if(allocated(lines_aud)) deallocate(lines_aud)
  if(allocated(lines_levelup)) deallocate(lines_levelup)
  if(allocated(lines_leveldown)) deallocate(lines_leveldown)
  if(allocated(lines_levelup_subsetindex)) deallocate(lines_levelup_subsetindex)
  if(allocated(lines_leveldown_subsetindex)) deallocate(lines_leveldown_subsetindex)
  if(allocated(lines_levelpop)) deallocate(lines_levelpop)
  if(allocated(gasvelocity)) deallocate(gasvelocity)
  if(allocated(lines_ray_levpop)) deallocate(lines_ray_levpop)
  if(allocated(lines_ray_temp)) deallocate(lines_ray_temp)
  if(allocated(lines_ray_turb)) deallocate(lines_ray_turb)
  if(allocated(lines_ray_doppler)) deallocate(lines_ray_doppler)
  if(allocated(lines_microturb)) deallocate(lines_microturb)
  if(allocated(gas_chemspec_numberdens)) deallocate(gas_chemspec_numberdens)
  if(allocated(lines_partition_temperature)) deallocate(lines_partition_temperature)
  if(allocated(lines_partition_function)) deallocate(lines_partition_function)
  if(allocated(lines_partition_ntemp)) deallocate(lines_partition_ntemp)
  if(allocated(active_lines)) deallocate(active_lines)
  if(allocated(active_nrlines)) deallocate(active_nrlines)
  if(allocated(active_levels)) deallocate(active_levels)
  if(allocated(active_levels_subsetindex)) deallocate(active_levels_subsetindex)
  if(allocated(active_nrlevels)) deallocate(active_nrlevels)
  if(allocated(active_dummy)) deallocate(active_dummy)
  lines_maxnrlevels = 0
  lines_maxnrlines  = 0
  lines_partition_ntempmax=0
  lines_nrlevels_subset_max = 0
  lines_nrlines_subset_max = 0
  lines_maxveloc = 0.d0
  lines_maxturbc = 0.d0
  lines_maxtempc = 0.d0
  lines_maxshift = 0.d0
  !
  ! Issue warning
  !
  if(lines_warn_lineleap) then
     write(stdo,*) '*****************************************************'
     write(stdo,*) '  WARNING from the line module:'
     write(stdo,*) '  Insufficient spatial resolution: doppler shift '
     write(stdo,*) '  between some adjacent cells larger than the '
     write(stdo,*) '  local line width.'
     write(stdo,*) '*****************************************************'
  endif
end subroutine lines_cleanup



!----------------------------------------------------------------------------
!                       READ ALL FOR THE LINES
!
! You can call this routine also if no line data are present. It 
! will then simply reset all line stuff to zero. In fact, by checking
! whether lines_nr_species.gt.0 you know whether the line data are
! present or not. 
!----------------------------------------------------------------------------
subroutine read_lines_all(action)
  implicit none
  integer :: action
  !
  ! Default
  !
  lines_pfunc_outofrange = .false.
  !
  ! Read the line fundamental data such as energy levels, degeneration
  ! coefficients, Einstein A coefficients etc.
  !
  ! This has already been initialized in userdef_prep_model for userdef lines.
  !
  if (lines_mode.ne.-10) then
    call read_linedata(action)
  endif
  !
  ! If lines_mode.ne.50 then we must have the number densities of
  ! the molecules because we will compute the populations internally,
  ! either via local prescriptions or full non-LTE. In principle, if
  ! the full non-LTE is done and these populations are to be re-read
  ! by RADMC-3D for ray-tracing, in that case it is not necessary to
  ! read these, but it can't hurt either. We also need the gas temperature,
  ! so read this as well.
  !
  if((lines_mode.ne.50).and.(lines_mode.ne.-10)) then 
     call read_molecules_numberdensities(action)
     call read_gas_temperature(action)
     lines_maxtempc = sqrt(2*kk*gastmax/(lines_umass_min))/cc
  endif
  !
  if(lines_mode.eq.-10) then
     call read_gas_temperature(action)
     lines_maxtempc = sqrt(2*kk*gastmax/(lines_umass_min))/cc
  endif
  !
  ! If abs(lines_mode).eq.1 then read the partition functions
  !
  if((lines_mode.eq.1).or.(lines_mode.eq.-1)) then
     call read_partition_function(action)
  endif
  !
  ! If lines_mode.ge.50, then read or compute the level populations if they
  ! are not yet in memory already.
  !
  if(lines_mode.ge.50) then 
     call read_levelpop(action)
  endif
  !
  ! The velocity field always has to be read if not yet done
  !
  call read_velocityfield(action)
  !
  ! Microturbulence is only read if a corresponding file is available
  !
  call read_microturbulence(action)
  !
  ! Now compute the maximum width of an observed line, i.e. how far
  ! away from some wavelength should I still look for lines that might
  ! contribute to this wavelength... For the stochastic parts (the
  ! microturbulence and the temperature) we include a factor 
  ! lines_widthmargin to see how far in the line wing we want to look.
  !
  lines_maxshift = lines_widthmargin * ( lines_maxtempc + lines_maxturbc ) + &
                   lines_maxveloc
  !
  ! Since we know the grid, we also know the largest length of a single
  ! ray. This means that we can already set up the 1-D formal transfer
  ! along a ray stuff.
  !
  ! ############## FUTURE: FOR NVIDIA GRAPHICS CARD #########
  !!!call lines_ray1d_init_raytrace(action)
  !
  call lines_serial_init_raytrace(action)
  !
end subroutine read_lines_all



!-------------------------------------------------------------------
!                    READ THE MAIN LINE DATA FILE
!-------------------------------------------------------------------
subroutine read_linedata(action)
  implicit none
  integer :: iformat,ierror,idum,ispec,idum1,idum2,ierr,ilevel,action
  character*80 :: ilinestring,jlinestring
  logical :: fex
  integer, allocatable :: subset_mode(:)
  !
  ! Action=0 means do nothing, action=1 means read if not yet read,
  ! action=2 means re-read.
  !
  if(action.eq.0) then
     return
  elseif(action.eq.1) then
     if(allocated(lines_nrlevels)) return
  endif
  !
  ! Clean up any remaining stuff
  !
  call lines_cleanup()
  !
  ! Set some default constants
  !
  !lines_widthmargin = 3.d0   ! Default should be from the main routine.
  lines_warn_lineleap = .false.
  !
  ! Open the lines.inp
  !
  inquire(file='lines.inp',exist=fex)
  if(.not.fex) then
     write(stdo,*) 'ERROR in line module: Could not find file lines.inp'
     stop
  endif
  !
  ! Message
  !
  write(stdo,*) 'Reading line data...'
  call flush(stdo)
  !
  ! Open file
  !
  open(unit=3,file='lines.inp',status='unknown')
  !
  ! Format number
  !
  read(3,*) iformat
  !
  ! Read how many species of molecules/atoms we want to read
  !
  read(3,*) lines_nr_species
  !
  ! Read the names of the input files and the max nr of 
  ! levels to take into account
  !
  ! First check if no strange number was entered
  !
  if((lines_nr_species.lt.1).or.(lines_nr_species.gt.100)) then
     write(stdo,*) 'ERROR in line module: lines_nr_species out of range'
     stop
  endif
  !
  ! Allocate some arrays
  !
  allocate(lines_nrlevels(1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_nrlevels(:).'
     stop 
  endif
  lines_nrlevels(:) = 0
  allocate(lines_nrlines(1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_nrlines(:).'
     stop 
  endif
  lines_nrlines(:) = 0
  allocate(lines_nrlevels_subset(1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_nrlevels_subset(:).'
     stop 
  endif
  lines_nrlevels_subset(:) = 0
  allocate(lines_nrlines_subset(1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_nrlines_subset(:).'
     stop 
  endif
  lines_nrlines_subset(:) = 0
  allocate(lines_umass(1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_umass(:).'
     stop 
  endif
  lines_umass(:) = 0.d0
  allocate(lines_speciesname(1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_speciesname(:).'
     stop 
  endif
  allocate(lines_filestyle(1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_filestyle(:).'
     stop 
  endif
  allocate(subset_mode(1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      subset_mode(:).'
     stop 
  endif
  subset_mode(:) = 0
  !
  ! Now read the species names, the input file format, the 
  ! way to select the nr of levels (idum1), the max nr of 
  ! levels (idum2). The last number means 'unspecified' if 0. 
  ! If idum1 is 0 then we select the levels only by their number limit.
  ! Example 1:
  !    co    leiden    0   0        ==> Read molecule_co.inp from
  !                                     the Leiden database and
  !                                     use all levels and lines.
  !    co    leiden    0   10       ==> Read molecule_co.inp from
  !                                     the Leiden database and
  !                                     use only levels 1 to 10
  !    co    leiden    1   10       ==> Read molecule_co.inp from
  !                                     the Leiden database and
  !                                     select 10 levels (which are
  !                                     specified further down).
  !
  do ispec=1,lines_nr_species
     !
     ! Read the info
     !
     read(3,*) ilinestring,jlinestring,idum1,idum2
     lines_speciesname(ispec)      = ilinestring
     lines_filestyle(ispec)        = jlinestring
     subset_mode(ispec)            = idum1
     lines_nrlevels_subset(ispec)  = idum2
     !
     ! Check if this file is present, and if so, check the nr of levels
     !
     call read_molecule(ispec,.false.,ierror)
     if(ierror.ne.0) then
        write(stdo,*) 'ERROR in line module: somehow cannot read species ',ispec
        stop
     endif
     !
     ! Check if the number of subset levels does not exceed the number of levels
     !
     if((lines_nrlevels_subset(ispec).le.0).or.                        &
        (lines_nrlevels_subset(ispec).gt.lines_nrlevels(ispec))) then
        lines_nrlevels_subset(ispec) = lines_nrlevels(ispec)
     endif
     !
  enddo
  !
  ! Find the maximum number of levels and lines available for each
  ! species
  !
  lines_maxnrlevels = 0
  lines_maxnrlines = 0
  do ispec=1,lines_nr_species
     lines_maxnrlevels = max(lines_maxnrlevels,lines_nrlevels(ispec))
     lines_maxnrlines = max(lines_maxnrlines,lines_nrlines(ispec))
  enddo
  !
  ! Find the maximum number of levels used for each species
  !
  lines_nrlevels_subset_max = 0
  do ispec=1,lines_nr_species
     lines_nrlevels_subset_max = max(lines_nrlevels_subset_max,     &
                                     lines_nrlevels_subset(ispec))
  enddo
  !
  ! Allocate the subset arrays. These arrays contain the indices of
  ! the levels/lines that are used. Please keep them in ascending
  ! order!!
  !
  allocate(lines_levels_subset(1:lines_maxnrlevels,lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_levels_subset(:,:).'
     stop 
  endif
  allocate(lines_lines_subset(1:lines_maxnrlines,lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_lines_subset(:,:).'
     stop 
  endif
  !
  ! Now set the subsets
  !
  do ispec=1,lines_nr_species
     if(subset_mode(ispec).eq.0) then
        !
        ! Just select the first nrlevels(ispec) levels
        !
        do ilevel=1,lines_nrlevels_subset(ispec)
           lines_levels_subset(ilevel,ispec) = ilevel
        enddo
     elseif(subset_mode(ispec).eq.1) then
        !
        ! Read a list of levels
        ! 
        read(1,*) (lines_levels_subset(ilevel,ispec), &
                   ilevel=1,lines_nrlevels_subset(ispec))
        !
        ! Check if all these are within range
        !
        do ilevel=1,lines_nrlevels_subset(ispec)
           if((lines_levels_subset(ilevel,ispec).lt.1).or.    &
              (lines_levels_subset(ilevel,ispec).gt.lines_nrlevels(ispec))) then
              write(stdo,*) 'ERROR in line module: subset indices for levels out of range'
              stop
           endif
        enddo
        !
        ! Check if they are in ascending order
        !
        do ilevel=2,lines_nrlevels_subset(ispec)
           if(lines_levels_subset(ilevel,ispec).le.lines_levels_subset(ilevel-1,ispec)) then
              write(stdo,*) 'ERROR: Subset of levels for species ',ispec
              write(stdo,*) '    is not in ascending order...'
              stop
           endif
        enddo
     else
        write(stdo,*) 'ERROR in line module: do not know subset_mode = ', &
             subset_mode(ispec)
        stop
     endif
  enddo
  !
  ! Close this file
  !
  close(3)
  !
  ! Now allocate the rest of the arrays
  !
  allocate(lines_nu0(1:lines_maxnrlines,1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_nu0(:).'
     stop 
  endif
  allocate(lines_aud(1:lines_maxnrlines,1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_aud(:).'
     stop 
  endif
  allocate(lines_levelup(1:lines_maxnrlines,1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_levelup(:).'
     stop 
  endif
  allocate(lines_leveldown(1:lines_maxnrlines,1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_leveldown(:).'
     stop 
  endif
  allocate(lines_levelup_subsetindex(1:lines_maxnrlines,1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_levelup_subsetindex(:).'
     stop 
  endif
  allocate(lines_leveldown_subsetindex(1:lines_maxnrlines,1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_leveldown_subsetindex(:).'
     stop 
  endif
  allocate(lines_level_energy_cminv(1:lines_maxnrlevels,1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_level_energy_cminv(:).'
     stop 
  endif
  allocate(lines_level_gdeg(1:lines_maxnrlevels,1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_level_gdeg(:).'
     stop 
  endif
  allocate(lines_level_name(1:lines_maxnrlevels,1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_level_name(:).'
     stop 
  endif
  !
  ! Now read each molecule/atom data
  !
  do ispec=1,lines_nr_species
     call read_molecule(ispec,.true.,ierror)
  enddo
  !
  ! Compute the total number of subset lines
  !
  lines_nrlines_subset_max=0
  do ispec=1,lines_nr_species
     lines_nrlines_subset_max = max(lines_nrlines_subset_max,   &
                                    lines_nrlines_subset(ispec))
  enddo
  !
  ! Allocate a local arrays for lines_serial_addto_jnu_alpnu() or
  ! lines_ray1d_addto_jnu_alpnu()
  !
  allocate(active_lines(lines_nrlines_subset_max,lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in line module: array active_lines() not allocatable'
     stop
  endif
  allocate(active_nrlines(lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in line module: array active_nrlines() not allocatable'
     stop
  endif
  allocate(active_levels(lines_nrlevels_subset_max,lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in line module: array active_levels() not allocatable'
     stop
  endif
  allocate(active_levels_subsetindex(lines_nrlevels_subset_max,lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in line module: array active_levels_subsetindex() not allocatable'
     stop
  endif
  allocate(active_nrlevels(lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in line module: array active_nrlevels() not allocatable'
     stop
  endif
  allocate(active_dummy(lines_maxnrlevels),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in line module: array active_dummy() not allocatable'
     stop
  endif
  !
  ! Pre-compute some stuff
  !
  lines_umass_min = 1.d99
  lines_umass_max = 0.d0
  do ispec=1,lines_nr_species
     if(lines_umass(ispec).lt.lines_umass_min) lines_umass_min = lines_umass(ispec)
     if(lines_umass(ispec).gt.lines_umass_max) lines_umass_max = lines_umass(ispec)
  enddo
  !
  ! Deallocate temporary array
  !
  if(allocated(subset_mode)) deallocate(subset_mode)
  !
end subroutine read_linedata


!-------------------------------------------------------------------
! READ A MOLECULAR/ATOMIC DATA FILE FROM ANY KIND OF SUPPORTED FILE
!-------------------------------------------------------------------
subroutine read_molecule(ispec,readdata,ierror)
  implicit none
  character*80 :: linestring
  integer :: ispec,ierror,iline,ilinsub
  logical :: readdata
  character*160 :: species
  !
  ! Call the reading routine for this specific database type
  !
  linestring = lines_filestyle(ispec)
  if(linestring(1:6).eq.'leiden') then
     call read_molecule_leiden(ispec,readdata,ierror)
     if(ierror.ne.0) then
        write(stdo,*) 'ERROR during line reading: failed. Aborting.'
        stop
     endif
  elseif(linestring(1:6).eq.'ratran') then
     write(stdo,*) 'RATRAN data format not yet implemented'
     stop
  else
     write(stdo,*) 'ERROR: Do not know file style ',linestring
     stop
  endif
  !
  ! Print out the line list, if requested
  !
  if(readdata.and.lines_make_linelist) then
     species  = lines_speciesname(ispec)
     write(stdo,*) 'Line list for molecule ',species(1:len_trim(species)),':'
     ilinsub=1
     do iline=1,lines_nrlines(ispec)
        if(lines_lines_subset(ilinsub,ispec).eq.iline) then
           write(stdo,*) 'Transition ',iline,': lambda = ',  &
                1d4*cc/lines_nu0(iline,ispec),' micron'
           ilinsub = ilinsub + 1
        else
           write(stdo,*) 'Transition ',iline,': lambda = ',  &
                1d4*cc/lines_nu0(iline,ispec),' micron [not included in model]'
        endif
     enddo
  endif
  !
end subroutine read_molecule


!-------------------------------------------------------------------
!    READ A MOLECULAR/ATOMIC DATA FILE FROM THE LEIDEN DATABASE
!-------------------------------------------------------------------
subroutine read_molecule_leiden(ispec,readdata,ierror)
  implicit none
  integer :: ispec,ierror,ierr
  logical :: readdata,fex,flag1,flag2
  character*160 filename,header,species
  integer :: nlevels,nlines,ilevel,iline,iup,idown,ilinesub,ilevelsub
  integer :: ilevel1,ilevel2
  integer, allocatable :: line_iup(:),line_idown(:)
  double precision, allocatable :: line_aud(:),line_nu0(:)
  double precision, allocatable :: level_energy_cminv(:),level_gdeg(:)
  character*256, allocatable :: level_name(:)
  character*256 :: name
  double precision :: e,g,dum1,dum2,dum3,aud,nu0,nucheck,error_linefreq 
  !
  ! Some check
  !
  if((.not.allocated(lines_aud)).and.readdata) then
     write(stdo,*) 'ERROR in line module: stuff not allocated...'
     stop
  endif
  if(ispec.gt.lines_nr_species) then
     write(stdo,*) 'ERROR in line module: ispec out of range...'
     stop
  endif
  !
  ! Initial setting
  !
  ierror = 0
  !
  ! Get species name
  !
  species  = lines_speciesname(ispec)
  !
  ! Message
  !
  if(readdata) then
     write(stdo,*) 'Reading line data of molecule ',species(1:len_trim(species)), &
          ' ...'
     call flush(stdo)
  endif
  !
  ! Find the file
  !
  filename = 'molecule_'//species(1:len_trim(species))//'.inp'
  inquire(file=filename,exist=fex)
  if(.not.fex) then
     ierror = 1
     return
  endif
  !
  ! Open the file
  !
  open(unit=1,file=filename)
  !
  ! Read the first dummy strings
  !      
  read(1,*) header
  read(1,*) header     ! The name of the molecule
  read(1,*) header
  !
  ! Read the molecular weight
  !
  read(1,*) lines_umass(ispec)
  lines_umass(ispec) = lines_umass(ispec)*mp
  !
  ! Read header
  !
  read(1,*) header
  !
  ! Read the number of levels
  !
  read(1,*) nlevels
  !
  ! Allocate the level energy and g array
  !
  allocate(level_energy_cminv(1:nlevels),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in line module: could not allocate temporary array level_energy_cminv'
     stop
  endif
  allocate(level_gdeg(1:nlevels),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in line module: could not allocate temporary array level_gdeg'
     stop
  endif
  allocate(level_name(1:nlevels),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in line module: could not allocate temporary array level_name'
     stop
  endif
  !
  ! Read header
  !
  read(1,*) header
  !
  ! Now do a loop over the levels and read their energies in cm^-1 and
  ! the degeneration parameters g.
  !
  ! NOTE: I have verified for at least one file of the Leiden LAMBDA
  !       database that the right light speed is used for the link between 
  !       the energies and the line positions: c=2.99792458d10 cm/s.
  !       Note that this value is *by definition* the light speed (it
  !       defines the length of the meter).
  !
  do ilevel=1,nlevels
     !
     ! Read the energy and degeneration
     !
     read(1,*) dum1,e,g,name
     !
     ! Store
     !
     level_energy_cminv(ilevel) = e
     level_gdeg(ilevel)         = g
     level_name(ilevel)         = name
  enddo
  !
  ! Read header
  !
  read(1,*) header
  !
  ! Now read the number of transitions
  !
  read(1,*) nlines
  !
  ! Allocate the line info
  !
  allocate(line_iup(1:nlines),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in line module: could not allocate temporary array line_iup'
     stop
  endif
  allocate(line_idown(1:nlines),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in line module: could not allocate temporary array line_idown'
     stop
  endif
  allocate(line_aud(1:nlines),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in line module: could not allocate temporary array line_aud'
     stop
  endif
  allocate(line_nu0(1:nlines),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in line module: could not allocate temporary array line_nu0'
     stop
  endif
  !
  ! Read header
  !
  read(1,*) header
  !
  ! Now loop over the lines
  !
  do iline=1,nlines
     !
     ! Read
     !
     read(1,*) dum1,iup,idown,aud,nu0,dum3
     !
     ! Store
     !
     line_iup(iline)   = iup
     line_idown(iline) = idown
     line_aud(iline)   = aud
     line_nu0(iline)   = nu0*1d9   ! The frequency is in GHz
  enddo
  !
  ! Close this file
  !
  close(1)
  !
  ! Some post-processing
  !
  lines_nrlevels(ispec) = nlevels
  lines_nrlines(ispec)  = nlines
  !
  ! OK, if we truly want to read the data, then do it now, else return
  ! with only the lines_nrlevels(ispec) and lines_nrlines(ispec) information.
  !
  if(readdata) then
     !
     ! Yes, we really want to read and store the entire data
     !
     ! First store the level info
     !
     do ilevel=1,lines_nrlevels(ispec)
        lines_level_energy_cminv(ilevel,ispec) = level_energy_cminv(ilevel)
        lines_level_gdeg(ilevel,ispec)         = level_gdeg(ilevel)
        lines_level_name(ilevel,ispec)         = level_name(ilevel)
     enddo
     !
     ! Then store the line info
     !
     do iline=1,lines_nrlines(ispec)
        lines_levelup(iline,ispec)   = line_iup(iline)
        lines_leveldown(iline,ispec) = line_idown(iline)
        lines_aud(iline,ispec)       = line_aud(iline)
        lines_nu0(iline,ispec)       = line_nu0(iline)
     enddo
     !
     ! Check the frequency of the transitions for consistency
     !     
     error_linefreq = 0.d0
     do iline=1,lines_nrlines(ispec)
        nucheck = cc * (                                                     &
             lines_level_energy_cminv(lines_levelup(iline,ispec),ispec) -    &
             lines_level_energy_cminv(lines_leveldown(iline,ispec),ispec) )
        if(abs(nucheck/line_nu0(iline)-1.d0).gt.error_linefreq) then
           error_linefreq = abs(nucheck/line_nu0(iline)-1.d0)
        endif
     enddo
     if(error_linefreq.gt.1d-7) then
        if(error_linefreq.gt.1d-3) then
           write(stdo,*) 'Warning: Level energies of molecule ',ispec,' are very'
           write(stdo,*) '   inaccurate. Relative error is: ',error_linefreq
        else
           write(stdo,*) 'Warning: Level energies of molecule ',ispec,' are slightly'
           write(stdo,*) '   inaccurate. Relative error is: ',error_linefreq
        endif
     endif
     !
     ! Now create the subset of lines for which both the upper and lower
     ! levels are contained in the subset of levels
     !
     ilinesub = 1
     do iline=1,lines_nrlines(ispec)
        flag1     = .false.
        flag2     = .false.
        ilevel    = lines_levelup(iline,ispec)
        do ilevelsub=1,lines_nrlevels_subset(ispec)
           if(lines_levels_subset(ilevelsub,ispec).eq.ilevel) then
              ilevel1=ilevelsub
              flag1=.true.
           endif
        enddo
        ilevel    = lines_leveldown(iline,ispec)
        do ilevelsub=1,lines_nrlevels_subset(ispec)
           if(lines_levels_subset(ilevelsub,ispec).eq.ilevel) then
              ilevel2=ilevelsub
              flag2=.true.
           endif
        enddo
        if(flag1.and.flag2) then
           lines_lines_subset(ilinesub,ispec)  = iline
           lines_levelup_subsetindex(iline,ispec)   = ilevel1
           lines_leveldown_subsetindex(iline,ispec) = ilevel2
           ilinesub = ilinesub+1
        endif
     enddo
     lines_nrlines_subset(ispec) = ilinesub-1
     !
     ! Check
     !
     if(lines_nrlines_subset(ispec).lt.1) then
        write(stdo,*) 'ERROR in line module, species ',ispec
        write(stdo,*) '      No lines found for the selected subset of levels'
        stop
     endif
     !
     ! If you like, here's a level diagram in ascii format
     !
     if(lines_show_pictograms.eq.1) then 
        write(*,*) 'Level Diagram for ispec=',ispec,':'
        call lines_pictlevels(ispec)
     endif
  endif
  !
  ! Deallocate temporary arrays
  !
  deallocate(level_energy_cminv)
  deallocate(level_gdeg)
  deallocate(level_name)
  deallocate(line_iup)
  deallocate(line_idown)
  deallocate(line_aud)
  deallocate(line_nu0)
  !
  ! Done.
  !
end subroutine read_molecule_leiden





!-------------------------------------------------------------------
!           MAKE GRAPHICAL PICTOGRAM OF LEVELS AND TRANSITIONS
!-------------------------------------------------------------------
subroutine lines_pictlevels(ispec)
  implicit none
  integer :: ispec
  integer :: i,k,nlinmax,istart,iend
  character*83 pict
  !
  istart=14
  nlinmax = min(lines_nrlines_subset(ispec),69)
  do k=lines_nrlevels_subset(ispec),1,-1
     write(pict(1:10),403) lines_level_energy_cminv(lines_levels_subset(k,ispec),ispec)
403  format(1F10.5)
     do i=11,istart
        pict(i:i) = ' '
     enddo
     iend = 1
     do i=1,nlinmax
        if((k.ge.lines_leveldown(lines_lines_subset(i,ispec),ispec)).and. &
           (k.le.lines_levelup(lines_lines_subset(i,ispec),ispec))) then
14         pict(i+istart:i+istart) = '|'
           iend = i+istart
        else
           pict(i+istart:i+istart) = ' '
        endif
     enddo
     do i=nlinmax+istart+1,83
        pict(i:i) = ' ' 
     enddo
     write(*,*) pict (1:iend+1)
  enddo
  !
end subroutine lines_pictlevels




!-------------------------------------------------------------------
!                   READ THE LEVEL POPULATIONS
!
! NOTE: These are the full level populations, i.e. in units of
!       1/cm^3. The molecular number density is already included
!       in these numbers. 
!-------------------------------------------------------------------
subroutine read_levelpop(action)
  implicit none
  character*80 :: filename1,filename2,filename3,filename,species
  integer :: ispec,ierr,iformat,nlevels,ilevel,nc,index,i,irec,n
  integer :: ntmp,itemp,action
  integer(kind=8) :: iiformat,reclen,reclend,nn,kk,nnlevels
  logical :: fex1,fex2,fex3
  double precision, allocatable :: thedata(:)
  integer(kind=8), allocatable :: subset(:)
  double precision, allocatable :: data(:,:)
  real, allocatable :: datas(:,:)
  !
  ! Consistency check
  !
  if(lines_mode.lt.0) stop 9351
  !
  ! Action=0 means do nothing, action=1 means read if not yet read,
  ! action=2 means re-read.
  !
  if(action.eq.0) then
     return
  elseif(action.eq.1) then
     if(allocated(lines_nrlevels)) return
  endif
  !
  ! Message
  !
  write(stdo,*) 'Reading level populations...'
  call flush(stdo)
  !
  ! Checks
  !
  if(lines_nrlevels_subset_max.le.0) then
     write(stdo,*) 'ERROR: level subset not yet specified while reading level populations.'
     stop
  endif
  !
  ! Remove any already existing array
  !
  if(allocated(lines_levelpop)) deallocate(lines_levelpop)
  !
  ! Create the level populations arrays
  !
  allocate(lines_levelpop(1:lines_nrlevels_subset_max,1:lines_nr_species,1:nrcells),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR: Could not allocate the lines_levelpop() array'
     stop
  endif
  !
  ! Now loop over all species
  !
  do ispec=1,lines_nr_species
     !
     ! Create the file name and find if the input file is present and if
     ! it is in text format (.inp) or unformatted (.uinp)
     !
     species   = lines_speciesname(ispec)
     filename1 = 'levelpop_'//species(1:len_trim(species))//'.inp'
     filename2 = 'levelpop_'//species(1:len_trim(species))//'.uinp'
     filename3 = 'levelpop_'//species(1:len_trim(species))//'.usinp'
     inquire(file=filename1,exist=fex1)
     inquire(file=filename2,exist=fex2)
     inquire(file=filename2,exist=fex3)
     if(fex1.and.(fex2.or.fex3)) then
        write(stdo,*) 'ERROR in line module: cannot have both formatted and'
        write(stdo,*) '   unformatted level population data files at the same time.'
        stop
     endif
     if((.not.fex1).and.(.not.fex2).and.(.not.fex3)) then
        write(stdo,*) 'ERROR in line module: for molecule species ',ispec, &
             ' could not find level population file.'
        stop
     endif
     if(fex1) then
        !
        ! Open formatted
        !
        open(unit=1,file=filename1)
        !
        ! Read format number
        !
        read(1,*) iformat
        !
        ! Read number of grid points
        !
        read(1,*) nc
        !
        ! Check if this is equal to that of the grid.inp
        !
        if(nc.ne.nrcellsinp) then
           write(stdo,*) 'ERROR in line module: nr of grid points specified in '
           write(stdo,*) '    ',filename1(1:len_trim(filename1)),' unequal to the'
           write(stdo,*) '     nr specified in the grid file.'
           stop
        endif
        !
        ! Read number of levels 
        !
        read(1,*) nlevels
        !
        ! Check if this equals the number of levels of the subset
        !
        if(nlevels.ne.lines_nrlevels_subset(ispec)) then
           write(stdo,*) 'ERROR in line module: nr of levels in the file '
           write(stdo,*) filename1(1:len_trim(filename1)),' unequal to the'
           write(stdo,*) 'number of selected levels of this molecule.'
           stop
        endif
        !
        ! Allocate temporary arrays
        !
        allocate(subset(1:nlevels),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR in line module: could not allocate subset()'
           stop
        endif
        allocate(thedata(1:nlevels),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR in line module: could not allocate data()'
           stop
        endif
        !
        ! Read the list of levels
        !
        read(1,*) (subset(ilevel),ilevel=1,nlevels)
        !
        ! Check if they are identical to those that were selected in
        ! the lines.inp file
        !
        do ilevel=1,nlevels
           if(subset(ilevel).ne.lines_levels_subset(ilevel,ispec)) then
              write(stdo,*) 'ERROR in line module: the subset of levels in the file'
              write(stdo,*) filename1(1:len_trim(filename1)),' unequal to those'
              write(stdo,*) 'selected in the line.inp file.'
              write(stdo,*) 'Here: ',subset(:)
              write(stdo,*) 'Line.inp: ',lines_levels_subset(:,ispec)
              stop
           endif
        enddo
        !
        ! Now do the big loop
        !
        call amr_resetcount()
        call amr_nextcell(index)
        do while(index.ge.0)
           read(1,*) (thedata(ilevel),ilevel=1,nlevels)
           if(index.gt.0) lines_levelpop(1:nlevels,ispec,index) = thedata(1:nlevels)
           call amr_nextcell(index)
        enddo
        !
        ! Close
        !
        close(1)
     else
        !
        ! Open unformatted
        !
        if(fex2.and.fex3) then
           write(stdo,*) 'ERROR: Found both ',trim(filename2),' and ',&
                trim(filename3)
           stop
        endif
        if(fex2) then
           filename = filename2
        else
           filename = filename3
        endif
        open(unit=1,file=filename,status='old',form='unformatted')
        read(1) iiformat,reclen
        if(iiformat.ne.1) then
           write(stdo,*) 'ERROR: Format number of '//TRIM(filename)//' is invalid/unknown.'
           write(stdo,*) 'Format number = ',iiformat
           write(stdo,*) 'Record length = ',reclen
           stop
        endif
        !
        ! Nr of cells
        ! 
        read(1) nn
        if(nn.ne.nrcellsinp) then
           write(stdo,*) 'ERROR: '//trim(filename)//' does not have same number'
           write(stdo,*) '       of cells as the grid.'
           write(stdo,*) nn,nrcellsinp
           stop
        endif
        !
        ! Nr of levels
        !
        read(1) nnlevels
        nlevels=nnlevels
        if(nlevels.ne.lines_nrlevels_subset(ispec)) then
           write(stdo,*) 'ERROR in line module: nr of levels in the file '
           write(stdo,*) trim(filename),' unequal to the'
           write(stdo,*) 'number of selected levels of this molecule.'
           stop
        endif
        !
        ! Read the level subset
        !
        allocate(subset(1:nlevels),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR in line module: could not allocate subset()'
           stop
        endif
        read(1) (subset(ilevel),ilevel=1,nlevels)
        !
        ! Check that we use the same subset in this file as in the 
        ! lines.inp
        !
        do ilevel=1,nlevels
           if(subset(ilevel).ne.lines_levels_subset(ilevel,ispec)) then
              write(stdo,*) 'ERROR in line module: the subset of levels in the file'
              write(stdo,*) filename2(1:len_trim(filename2)),' unequal to those'
              write(stdo,*) 'selected in the line.inp file.'
              write(stdo,*) 'Here: ',subset(:)
              write(stdo,*) 'Line.inp: ',lines_levels_subset(:,ispec)
              stop
           endif
        enddo
        !
        ! Prepare for the record-based reading. NOTE: This is a nasty
        ! feature of FORTRAN 90/95: It cannot read purely binary data.
        ! It uses instead "record-based unformatted data", which is the
        ! last remaining stone-age fortran feature still in F90/95. In
        ! Fortran 2003 this is fixed: There a "stream-based unformatted"
        ! file format is possible, which is similar to the unformatted
        ! data of C or C++. In principle many F90/95 compilers already
        ! have this F03 feature implemented, but for portability we do
        ! not yet rely on it. So instead we divide the main data into
        ! "records" of a (preferably) large size, though the size cannot
        ! exceed 256^2 = 65536 bytes. The record length should be given
        ! at the start of the file: the reclen variable (which is a
        ! long integer = 8 bytes, as all other integers). The user must
        ! make sure that this is precisely the number of variables per
        ! record times the length of the float. For .uinp files (which
        ! are double precision) the length per variable is 8 bytes, and
        ! for .usinp files (which are single precision) the length per
        ! variable is 4 bytes. The "reclend" variable then gives for 
        ! how many cells there is information per record.
        !
        ! In this particular case we organize the populations of all 
        ! levels of one or more cells into a single record. That is
        ! why the "reclend = reclen/(8*nlevels)" for doubles.
        ! NOTE: Since the record length is max 65536, this means that
        ! the max nr of cells in each record is now also nlevels x less.
        !
        if(allocated(data)) deallocate(data)
        if(allocated(datas)) deallocate(datas)
        if(fex2) then
           reclend=reclen/(8*nlevels)
           allocate(data(1:nlevels,1:reclend),STAT=ierr)
        else
           reclend=reclen/(4*nlevels)
           allocate(datas(1:nlevels,1:reclend),STAT=ierr)
           allocate(data(1:nlevels,1:reclend),STAT=ierr)
        endif
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR: Could not allocate data array'
           stop
        endif
        call amr_resetcount()
        call amr_nextcell(index)
        do irec=1,(nrcellsinp-1)/reclend+1
           if(fex2) then
              read(1) data
           else
              read(1) datas
              data(:,:) = datas(:,:)
           endif
           i = 1
           do while((index.ge.0).and.(i.le.reclend))
              if(index.gt.0) lines_levelpop(1:nlevels,ispec,index) = data(1:nlevels,i)
              i     = i + 1
              call amr_nextcell(index)
           enddo
        enddo
        close(1)
     endif
  enddo ! Loop over species
  !
  ! Deallocate
  !
  if(allocated(subset)) deallocate(subset)
  if(allocated(data)) deallocate(data)
  if(allocated(datas)) deallocate(datas)
  if(allocated(thedata)) deallocate(thedata)
  !
end subroutine read_levelpop



!-------------------------------------------------------------------
!                 READ THE PARTITION FUNCTION
!
! Read, or generate internally, the partition functions for each of
! the species.
!
! Note: If lines_mode.ge.1 and there exists a level population file
!       for some species (or for all species), then of course the
!       partition function is not required. Otherwise it is required.
!       We will check for this.
!-------------------------------------------------------------------
subroutine read_partition_function(action)
  implicit none
  integer :: action
  integer :: ispec,ntmp,iformat,itmp,ilevel,itemp,ierr
  double precision :: dum1,dum2,lpop,gast,dummy,dendivk,const
  character*80 :: filename1,filename2,species
  logical :: fex1,fex2
  parameter(const=hh*cc/kk)
  !
  ! Action=0 means do nothing, action=1 means read if not yet read,
  ! action=2 means re-read.
  !
  if(action.eq.0) then
     return
  elseif(action.eq.1) then
     if(allocated(lines_partition_temperature)) return
  endif
  !
  ! Message
  !
  write(stdo,*) 'Reading or computing partition functions...'
  call flush(stdo)
  !
  ! Do stupidity checks
  !
  if(lines_nr_species.le.0) stop 8010
  if(.not.allocated(lines_speciesname)) stop 8011
  if(.not.allocated(gastemp)) then
     write(stdo,*) 'ERROR in line module: wanting to compute/read the partition function, but'
     write(stdo,*) '     the gas temperature has not been read into memory yet.'
     stop
  endif
  if(.not.allocated(gas_chemspec_numberdens)) then
     write(stdo,*) 'ERROR in line module: while compute/read the partition function:'
     write(stdo,*) '     The number density of the molecule(s) has not been read into memory yet.'
     write(stdo,*) '     While they are not required for the partition function, '
     write(stdo,*) '     they will be necessary later.'
     stop 
  endif
  if(.not.allocated(lines_level_energy_cminv)) stop 8014
  !
  ! First we need to find out the maximum number of temperature sampling
  ! points for the partition function. 
  !
  lines_partition_ntempmax = 0
  do ispec=1,lines_nr_species
     species   = lines_speciesname(ispec)
     filename2  = 'partitionfunction_'//species(1:len_trim(species))//'.inp'
     inquire(file=filename2,exist=fex2)
     if(fex2) then
        !
        ! A file for the partition function exists for this species. So
        ! get the number of temperature sampling points from this file.
        !
        open(unit=2,file=filename2)
        read(2,*) iformat
        read(2,*) ntmp
        close(2)
        if(ntmp.gt.lines_partition_ntempmax) then
           lines_partition_ntempmax = ntmp
        endif
     else
        !
        ! No file exists for the partition function for this species.
        ! This automatically means that we are going to compute the
        ! partition function internally. For this we need a number of
        ! sampling points given by lines_partition_ntempint.
        !
        if(lines_partition_ntempint.gt.lines_partition_ntempmax) then
           lines_partition_ntempmax = lines_partition_ntempint
        endif
     endif
  enddo
  !
  ! Deallocate stuff if necessary
  !
  if(allocated(lines_partition_ntemp)) deallocate(lines_partition_ntemp)
  if(allocated(lines_partition_temperature)) deallocate(lines_partition_temperature)
  if(allocated(lines_partition_function)) deallocate(lines_partition_function)
  !
  ! Now allocate the arrays for the partition function
  !
  allocate(lines_partition_ntemp(1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR: Could not allocate the lines_partition_ntemp() array'
     stop
  endif
  allocate(lines_partition_temperature(1:lines_partition_ntempmax,1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR: Could not allocate the lines_partition_temperature() array'
     stop
  endif
  allocate(lines_partition_function(1:lines_partition_ntempmax,1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR: Could not allocate the lines_partition_function() array'
     stop
  endif
  !
  ! Now get the partition function for all species, either by reading
  ! them from a file or by computing them on-the-fly
  !
  do ispec=1,lines_nr_species
     !
     ! Get the name of the species
     !
     species  = lines_speciesname(ispec)
     !
     ! Check if there is a partition function 
     !
     filename1 = 'partitionfunction_'//species(1:len_trim(species))//'.inp'
     inquire(file=filename1,exist=fex1)
     if(fex1) then
        !
        ! Now open the file for the partition function
        !
        open(unit=1,file=filename1)
        read(1,*) iformat
        read(1,*) ntmp
        lines_partition_ntemp(ispec) = ntmp
        do itemp=1,ntmp
           read(1,*) dum1,dum2
           lines_partition_temperature(itemp,ispec) = dum1
           lines_partition_function(itemp,ispec)    = dum2
        enddo
        close(1)
     else
        !
        ! No file available, so we create the partition function internally.
        !
        lines_partition_ntemp(ispec) = lines_partition_ntempint
        !
        ! Note: hh * cc / kk = 1.4387519
        ! NOTE: Now we use the constants_module.f90 for this. 
        !
        ! Loop over temperatures
        !
        do itemp=1,lines_partition_ntemp(ispec)
           !
           ! Create a temperature point
           !
           lines_partition_temperature(itemp,ispec) =                     &
              lines_partition_temp0 *                                     &
              (lines_partition_temp1/lines_partition_temp0)**             &
              (dble(itemp-1.d0)/dble(lines_partition_ntemp(ispec)-1.d0))
           !
           ! Now compute the partition function based on all the available
           ! levels for this molecule. 
           !
           ! Note: In particular for high temperatures this internal
           ! calculation may fail.
           !
           lpop  = lines_level_gdeg(1,ispec)
           dummy = lpop
           gast  = lines_partition_temperature(itemp,ispec)
           do ilevel=2,lines_nrlevels(ispec)
              dendivk = const *                                           &
                        ( lines_level_energy_cminv(ilevel,ispec) -        &
                          lines_level_energy_cminv(ilevel-1,ispec) )
              lpop    = lpop * exp(-dendivk/gast) *                       &
                        lines_level_gdeg(ilevel,ispec) /                  &
                        lines_level_gdeg(ilevel-1,ispec)
              dummy   = dummy + lpop
           enddo
           lines_partition_function(itemp,ispec) = dummy
        enddo
     endif
  enddo ! Loop over species
end subroutine read_partition_function


! *** AT THE MOMENT NOT BEING USED ***
! !-------------------------------------------------------------------
! !            SET ALL THE LEVEL POPULATIONS ACCORDING TO LTE
! !-------------------------------------------------------------------
! subroutine lines_compute_levelpop_lte(action)
!   implicit none
!   integer :: action,ierr,icell,ispec,index
!   !
!   ! Check
!   !
!   if(lines_mode.ne.1) then
!      write(stdo,*) 'ERROR: Cannot make LTE populations if lines_mode.ne.1.'
!      stop
!   endif
!   !
!   ! Action=0 means do nothing, action=1 means read if not yet read,
!   ! action=2 means re-read.
!   !
!   if(action.eq.0) then
!      return
!   elseif(action.eq.1) then
!      if(allocated(lines_levelpop)) return
!   endif
!   !
!   ! Message
!   !
!   write(stdo,*) 'Computing LTE level populations...'
!   call flush(stdo)
!   !
!   ! Checks
!   !
!   if(.not.allocated(lines_partition_temperature)) then
!      write(stdo,*) 'ERROR: wanting to compute level populations '
!      write(stdo,*) '       according to LTE prescription, but'
!      write(stdo,*) '       partition function array not allocated.'
!      stop 
!   endif
!   if(.not.allocated(gastemp)) then
!      write(stdo,*) 'ERROR: Wanting to compute level populations '
!      write(stdo,*) '       according to LTE prescription, but'
!      write(stdo,*) '       gastemp array not allocated.'
!      stop
!   endif
!   if(.not.allocated(gas_chemspec_numberdens)) then
!      write(stdo,*) 'ERROR: Wanting to compute level populations '
!      write(stdo,*) '       according to LTE prescription, but'
!      write(stdo,*) '       gas_chemspec_numberdens array not allocated.'
!      stop
!   endif
!   if(.not.allocated(cellindex)) then
!      write(stdo,*) 'ERROR: Wanting to compute level populations '
!      write(stdo,*) '       according to LTE prescription, but'
!      write(stdo,*) '       cellindex array not allocated.'
!      stop
!   endif
!   !
!   ! Deallocate array
!   !
!   if(allocated(lines_levelpop)) deallocate(lines_levelpop)
!   !
!   ! Reallocate array
!   !
!   allocate(lines_levelpop(1:lines_nrlevels_subset_max,1:lines_nr_species,1:nrcells),STAT=ierr)
!   if(ierr.ne.0) then
!      write(stdo,*) 'ERROR: Could not allocate the lines_levelpop() array'
!      stop
!   endif
!   !
!   ! Now loop over all locations on the grid and all species
!   !
!   do icell=1,nrcells
!      index = cellindex(icell)
!      do ispec=1,lines_nr_species
!         call lines_compute_ltepop(lines_nrlevels_subset(ispec),     &
!                  lines_levels_subset(:,ispec),###HERE DINDGEN###,ispec,gastemp(index), &
!                  gas_chemspec_numberdens(ispec,index),              &
!                  lines_levelpop(:,ispec,index))
!      enddo
!   enddo
! end subroutine lines_compute_levelpop_lte


!-------------------------------------------------------------------
!              MAKE LTE LEVEL POPULATIONS AT ONE POINT
!
! Note: Here you can specify a local subset of levels, which does
!       not necessarily have to be the same as the subset specified
!       in lines_levels_subset(:,:). This is useful for on-the-fly
!       computation of level populations, where you would not want
!       to compute level populations that are not involved in this
!       particular spectrum or image. 
!-------------------------------------------------------------------
subroutine lines_compute_ltepop(nlev,ilevsubset,ilevsubsetidx, &
                                ispec,gast,nrdens,levelpop)
  implicit none
  integer :: nlev,ispec,itemp,ilevel
  integer :: ilevsubset(1:nlev),ilevsubsetidx(1:nlev)
  double precision :: levelpop(1:nlev),gast,eps,pfunc,dendivk,nrdens,dummy
  !
  ! If no levels are selected, then return 
  !
  if(nlev.eq.0) return
  !
  ! Find the location of the current temperature in the parition function table
  !
  call hunt(lines_partition_temperature(:,ispec),               &
            lines_partition_ntemp(ispec),gast,itemp)
  if(itemp.ge.lines_partition_ntemp(ispec)) then
     write(stdo,*) 'ERROR: Temperature out of range of partition function '
     write(stdo,*) '       for ispec=',ispec,' Temperature = ',gast
     !write(stdo,*) '       Smallest temperature in table = ',   &
     !     lines_partition_temperature(1,ispec)
     write(stdo,*) '       Largest temperature in table = ',   &
          lines_partition_temperature(lines_partition_ntemp(ispec),ispec)
     stop
  endif
  if(itemp.ge.1) then
     eps = (gast-lines_partition_temperature(itemp,ispec)) /            &
          (lines_partition_temperature(itemp+1,ispec)-                  &
          lines_partition_temperature(itemp,ispec))
  else
     itemp = 1
     eps   = 0.d0
     lines_pfunc_outofrange = .true.
  endif
  if((eps.lt.0.).or.(eps.gt.1.)) stop 7729
  !
  ! Construct the partition function by linear interpolation in the table
  !
  pfunc = (1-eps)*lines_partition_function(itemp,ispec) +            &
              eps*lines_partition_function(itemp+1,ispec)
  !
  ! Now construct all the populations according to LTE
  ! Note that c*h/k=1.4387519 in CGS units.
  !
  dummy = nrdens / pfunc
  do ilevel=1,nlev
     dendivk = 1.4387519 *                                           &
             ( lines_level_energy_cminv(ilevsubset(ilevel),ispec)    &
             - lines_level_energy_cminv(1,ispec) )
     levelpop(ilevsubsetidx(ilevel)) = dummy * exp(-dendivk/gast) *  &
             lines_level_gdeg(ilevsubset(ilevel),ispec)
  enddo
  !
end subroutine lines_compute_ltepop



!-------------------------------------------------------------------
!            READ THE NUMBER DENSITIES OF TE MOLECULE(S)
!-------------------------------------------------------------------
subroutine read_molecules_numberdensities(action)
  implicit none
  integer :: action,ispec,index,i,ierr,iformat,irec,n
  integer(kind=8) :: iiformat,reclen,reclend,nn,kk
  character*80 :: filename1,filename2,filename3,filename,species
  real, allocatable :: datas(:)
  double precision, allocatable :: data(:)
  double precision :: dummy
  logical :: fex1,fex2,fex3
  !
  ! Action=0 means do nothing, action=1 means read if not yet read,
  ! action=2 means re-read.
  !
  if(action.eq.0) then
     return
  elseif(action.eq.1) then
     if(allocated(gas_chemspec_numberdens)) return
  endif
  !
  ! Message
  !
  write(stdo,*) 'Reading molecular/atomic number densities...'
  call flush(stdo)
  !
  ! Deallocate stuff if required
  !
  if(allocated(gas_chemspec_numberdens)) deallocate(gas_chemspec_numberdens)
  !
  ! Allocate memory for the number density of each of the species.
  !
  ! NOTE: We may in the future save some memory by not allocating this
  !       if none of the species have a number density file. For now
  !       we will not yet optimize to that level, for the sake of simplicity
  !
  allocate(gas_chemspec_numberdens(1:lines_nr_species,1:nrcells),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR: Could not allocate the gas_chemspec_numberdens() array'
     stop
  endif
  !
  ! Now loop over all species
  !
  do ispec=1,lines_nr_species
     !
     ! Get the name of the species
     !
     species  = lines_speciesname(ispec)
     !
     ! Check the file names for the number densities of this species
     ! If no such file is available, then there MUST be a file with
     ! the level populations. If a file is available, then we have
     ! the possibility to calculate e.g. LTE populations internally.
     !
     filename1 = 'numberdens_'//species(1:len_trim(species))//'.inp'
     filename2 = 'numberdens_'//species(1:len_trim(species))//'.uinp'
     filename3 = 'numberdens_'//species(1:len_trim(species))//'.usinp'
     inquire(file=filename1,exist=fex1)
     inquire(file=filename2,exist=fex2)
     inquire(file=filename3,exist=fex3)
     if(fex1.and.(fex2.or.fex3)) then
        write(stdo,*) 'ERROR in line module: cannot have both formatted and'
        write(stdo,*) '   unformatted number density data files at the same time.'
        stop
     endif
     if((.not.fex1).and.(.not.fex2).and.(.not.fex3)) then
        write(stdo,*) 'ERROR in line module: cannot file containing the'
        write(stdo,*) '   number density for molecule ',ispec
        stop
     endif
     !
     ! Message
     !
     write(stdo,*) 'Reading gas species number densities...'
     call flush(stdo)
     !
     ! Read
     !
     if(fex1) then
        !
        ! Formatted input for the number density of this species
        !
        open(unit=1,file=filename1,status='old')
        read(1,*) iformat
        if(iformat.ne.1) then
           write(stdo,*) 'ERROR: Format number of '//TRIM(filename1)//' is invalid/unknown.'
           stop
        endif
        read(1,*) n
        if(n.ne.nrcellsinp) then
           write(stdo,*) 'ERROR: '//trim(filename1)//' does not have same number'
           write(stdo,*) '       of cells as the grid.'
           write(stdo,*) n,nrcellsinp
           stop
        endif
        call amr_resetcount()
        call amr_nextcell(index)
        do while(index.ge.0)
           read(1,*) dummy 
           if(index.gt.0) gas_chemspec_numberdens(ispec,index) = max(dummy,1d-99)
           call amr_nextcell(index)
        enddo
        close(1)
     else
        !
        ! Unformatted input for the number density of this species
        !
        if(fex2.and.fex3) then
           write(stdo,*) 'ERROR: Found both ',trim(filename2),' and ',&
                trim(filename3)
           stop
        endif
        if(fex2) then
           filename = filename2
        else
           filename = filename3
        endif
        open(unit=1,file=filename,status='old',form='unformatted')
        read(1) iiformat,reclen
        if(iiformat.ne.1) then
           write(stdo,*) 'ERROR: Format number of '//TRIM(filename)//' is invalid/unknown.'
           write(stdo,*) 'Format number = ',iiformat
           write(stdo,*) 'Record length = ',reclen
           stop
        endif
        read(1) nn
        if(nn.ne.nrcellsinp) then
           write(stdo,*) 'ERROR: '//trim(filename)//' does not have same number'
           write(stdo,*) '       of cells as the grid.'
           write(stdo,*) nn,nrcellsinp
           stop
        endif
        !
        ! Prepare for the record-based reading. NOTE: This is a nasty
        ! feature of FORTRAN 90/95: It cannot read purely binary data.
        ! It uses instead "record-based unformatted data", which is the
        ! last remaining stone-age fortran feature still in F90/95. In
        ! Fortran 2003 this is fixed: There a "stream-based unformatted"
        ! file format is possible, which is similar to the unformatted
        ! data of C or C++. In principle many F90/95 compilers already
        ! have this F03 feature implemented, but for portability we do
        ! not yet rely on it. So instead we divide the main data into
        ! "records" of a (preferably) large size, though the size cannot
        ! exceed 256^2 = 65536 bytes. The record length should be given
        ! at the start of the file: the reclen variable (which is a
        ! long integer = 8 bytes, as all other integers). The user must
        ! make sure that this is precisely the number of variables per
        ! record times the length of the float. For .uinp files (which
        ! are double precision) the length per variable is 8 bytes, and
        ! for .usinp files (which are single precision) the length per
        ! variable is 4 bytes. The "reclend" variable then gives for 
        ! how many cells there is information per record.
        !
        if(allocated(data)) deallocate(data)
        if(allocated(datas)) deallocate(datas)
        if(fex2) then
           reclend=reclen/8
           allocate(data(1:reclend),STAT=ierr)
        else
           reclend=reclen/4
           allocate(datas(1:reclend),STAT=ierr)
           allocate(data(1:reclend),STAT=ierr)
        endif
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR: Could not allocate data array'
           stop
        endif
        call amr_resetcount()
        call amr_nextcell(index)
        do irec=1,(nrcellsinp-1)/reclend+1
           if(fex2) then
              read(1) data
           else
              read(1) datas
              data(:) = datas(:)
           endif
           i = 1
           do while((index.ge.0).and.(i.le.reclend))
              if(index.gt.0) gas_chemspec_numberdens(ispec,index) = max(data(i),1d-99)
              i     = i + 1
              call amr_nextcell(index)
           enddo
        enddo
        close(1)
     endif
     if(allocated(data)) deallocate(data)
     if(allocated(datas)) deallocate(datas)
  enddo
end subroutine read_molecules_numberdensities


!-------------------------------------------------------------------
!                    READ THE VELOCITY FIELD
!-------------------------------------------------------------------
subroutine read_velocityfield(action)
  implicit none
  character*80 :: filename1,filename2,filename3,filename
  integer :: ispec,ierr,iformat,nc,index,action,icell,i,irec
  integer(kind=8) :: iiformat,reclen,reclend,nn,kk
  logical :: fex1,fex2,fex3
  double precision :: vel(1:3)
  double precision, allocatable :: data(:,:)
  real, allocatable :: datas(:,:)
  double precision :: dummy
  !
  ! Action=0 means do nothing, action=1 means read if not yet read,
  ! action=2 means re-read.
  !
  if(action.eq.0) then
     return
  elseif(action.eq.1) then
     if(allocated(gasvelocity)) then
        !
        ! The gas velocity is already determined. It might be
        ! that also the lines_maxveloc has been calculated from
        ! this, but we don't know for sure. So let's calculate
        ! it here to make sure lines_maxveloc is up-to-date. It
        ! can be important if the gas velocity is created 
        ! using the userdef module.
        !
        ! Thanks, Rainer Rolffs, for pointing this out!
        ! 02.01.2011
        !
        lines_maxveloc = 0.d0
        do icell=1,nrcells
           index = cellindex(icell)
           dummy = gasvelocity(1,index)**2 + gasvelocity(2,index)**2 + gasvelocity(3,index)**2 
           dummy = sqrt(dummy)/cc
           if(dummy.gt.lines_maxveloc) then
              lines_maxveloc = dummy
           endif
        enddo
        return
     endif
  endif
  !
  ! Message
  !
  write(stdo,*) 'Reading velocity field...'
  call flush(stdo)
  !
  ! First create the gas velocity arrays
  !
  if(allocated(gasvelocity)) deallocate(gasvelocity)
  allocate(gasvelocity(1:3,1:nrcells),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR: Could not allocate the gasvelocity() array'
     stop
  endif
  !
  ! Create the file name and find if the input file is present and if
  ! it is in text format (.inp) or unformatted (.uinp)
  !
  filename1 = 'gas_velocity.inp'
  filename2 = 'gas_velocity.uinp'
  filename3 = 'gas_velocity.usinp'
  inquire(file=filename1,exist=fex1)
  inquire(file=filename2,exist=fex2)
  inquire(file=filename3,exist=fex3)
  if(fex1.and.(fex2.or.fex3)) then
     write(stdo,*) 'ERROR in line module: cannot have both formatted and'
     write(stdo,*) '   unformatted gas velocity data files at the same time.'
     stop
  endif
  if((.not.fex1).and.(.not.fex2).and.(.not.fex3)) then
     write(stdo,*) 'No gas velocity input file found. Assuming zero gas velocity...'
     gasvelocity(:,:) = 0.d0
     return
  endif
  !
  ! Switch formatted/unformatted
  !
  if(fex1) then
     !
     ! Open formatted
     !
     open(unit=1,file=filename1)
     !
     ! Read format number
     !
     read(1,*) iformat
     !
     ! Read number of grid points
     !
     read(1,*) nc
     !
     ! Check if this is equal to that of the grid.inp
     !
     if(nc.ne.nrcellsinp) then
        write(stdo,*) 'ERROR in line module: nr of grid points specified in '
        write(stdo,*) '    ',filename1(1:len_trim(filename1)),' unequal to the'
        write(stdo,*) '     nr specified in the grid file.'
        stop
     endif
     !
     ! Now do the big loop
     !
     call amr_resetcount()
     call amr_nextcell(index)
     do while(index.ge.0)
        read(1,*) vel(1),vel(2),vel(3)
        if(index.gt.0) gasvelocity(1:3,index) = vel(1:3)
        call amr_nextcell(index)
     enddo
     !
     ! Close
     !
     close(1)
  else
     !
     ! Open unformatted
     !
     if(fex2.and.fex3) then
        write(stdo,*) 'ERROR: Found both ',trim(filename2),' and ',&
             trim(filename3)
        stop
     endif
     if(fex2) then
        filename = filename2
     else
        filename = filename3
     endif
     open(unit=1,file=filename,form='unformatted')
     !
     ! Read format number
     !
     read(1) iiformat,reclen
     if(iiformat.ne.1) then
        write(stdo,*) 'ERROR: Format number of '//TRIM(filename)//' is invalid/unknown.'
        write(stdo,*) 'Format number = ',iiformat
        write(stdo,*) 'Record length = ',reclen
        stop
     endif
     read(1) nn
     if(nn.ne.nrcellsinp) then
        write(stdo,*) 'ERROR: '//trim(filename)//' does not have same number'
        write(stdo,*) '       of cells as the grid.'
        write(stdo,*) nn,nrcellsinp
        stop
     endif
     !
     ! Prepare for the record-based reading. NOTE: This is a nasty
     ! feature of FORTRAN 90/95: It cannot read purely binary data.
     ! It uses instead "record-based unformatted data", which is the
     ! last remaining stone-age fortran feature still in F90/95. In
     ! Fortran 2003 this is fixed: There a "stream-based unformatted"
     ! file format is possible, which is similar to the unformatted
     ! data of C or C++. In principle many F90/95 compilers already
     ! have this F03 feature implemented, but for portability we do
     ! not yet rely on it. So instead we divide the main data into
     ! "records" of a (preferably) large size, though the size cannot
     ! exceed 256^2 = 65536 bytes. The record length should be given
     ! at the start of the file: the reclen variable (which is a
     ! long integer = 8 bytes, as all other integers). The user must
     ! make sure that this is precisely the number of variables per
     ! record times the length of the float. For .uinp files (which
     ! are double precision) the length per variable is 8 bytes, and
     ! for .usinp files (which are single precision) the length per
     ! variable is 4 bytes. The "reclend" variable then gives for 
     ! how many cells there is information per record.
     !
     ! In this particular case we have three floats per cell: the
     ! three velocity components. Hence the (8*3) or (4*3) below.
     ! NOTE: Since the record length is max 65536, this means that
     ! the max nr of cells in each record is now also 3x less.
     !
     if(allocated(data)) deallocate(data)
     if(allocated(datas)) deallocate(datas)
     if(fex2) then
        reclend=reclen/(8*3)
        allocate(data(1:3,1:reclend),STAT=ierr)
     else
        reclend=reclen/(4*3)
        allocate(datas(1:3,1:reclend),STAT=ierr)
        allocate(data(1:3,1:reclend),STAT=ierr)
     endif
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate data array'
        stop
     endif
     !
     ! Now do the big loop
     !
     call amr_resetcount()
     call amr_nextcell(index)
     do irec=1,(nrcellsinp-1)/reclend+1
        if(fex2) then
           read(1) data
        else
           read(1) datas
           data(:,:) = datas(:,:)
        endif
        i = 1
        do while((index.ge.0).and.(i.le.reclend))
           if(index.gt.0) gasvelocity(1:3,index) = data(1:3,i)
           i     = i + 1
           call amr_nextcell(index)
        enddo
     enddo
     !
     ! Close
     !
     close(1)
     if(allocated(data)) deallocate(data)
     if(allocated(datas)) deallocate(datas)
  endif
  !
  ! Determine the maximum absolute velocity. This is useful for when we
  ! have to figure out which lines may contribute to a certain wavelength.
  !
  lines_maxveloc = 0.d0
  do icell=1,nrcells
     index = cellindex(icell)
     dummy = gasvelocity(1,index)**2 + gasvelocity(2,index)**2 + gasvelocity(3,index)**2 
     dummy = sqrt(dummy)/cc
     if(dummy.gt.lines_maxveloc) then
        lines_maxveloc = dummy
     endif
  enddo
  !
end subroutine read_velocityfield


!-------------------------------------------------------------------
!       READ THE LOCAL UNRESOLVED MICRO TURBULENCE (OPTIONAL)
!-------------------------------------------------------------------
subroutine read_microturbulence(action)
  implicit none
  character*80 :: filename1,filename2,filename3,filename
  integer(kind=8) :: iiformat,reclen,reclend,nn,kk
  integer :: ierr,iformat,nc,icell,index,action,i,irec
  real, allocatable :: datas(:)
  double precision, allocatable :: data(:)
  logical :: fex1,fex2,fex3
  real :: dummy
  !
  ! Action=0 means do nothing, action=1 means read if not yet read,
  ! action=2 means re-read.
  !
  if(action.eq.0) then
     return
  elseif(action.eq.1) then
     if(allocated(lines_microturb)) then
        !
        ! The microturbulence is already determined. It might be
        ! that also the lines_maxturbc has been calculated from
        ! this, but we don't know for sure. So let's calculate
        ! it here to make sure lines_maxturbc is up-to-date. It
        ! can be important if the microturbulence is created 
        ! using the userdef module.
        !
        ! Thanks, Rainer Rolffs, for pointing this out!
        ! 02.01.2011
        !
        lines_maxturbc = 0.d0
        do icell=1,nrcells
           index = cellindex(icell)
           if(lines_microturb(index).gt.lines_maxturbc) then
              lines_maxturbc = lines_microturb(index)
           endif
        enddo
        lines_maxturbc = lines_maxturbc / cc
        return
     endif
  endif
  !
  ! Deallocate any possible present array
  !
  if(allocated(lines_microturb)) deallocate(lines_microturb)
  !
  ! Create the file name and find if the input file is present and if
  ! it is in text format (.inp) or unformatted (.uinp)
  !
  filename1 = 'microturbulence.inp'
  filename2 = 'microturbulence.uinp'
  filename3 = 'microturbulence.usinp'
  inquire(file=filename1,exist=fex1)
  inquire(file=filename2,exist=fex2)
  inquire(file=filename3,exist=fex3)
  if(fex1.and.(fex2.or.fex3)) then
     write(stdo,*) 'ERROR in line module: cannot have both formatted and'
     write(stdo,*) '   unformatted microturbulence data files at the same time.'
     stop
  endif
  !
  ! If no microturbulence file is present, then do not use microturbulence
  !
  if((.not.fex1).and.(.not.fex2).and.(.not.fex3)) then
     write(stdo,*) 'No microturbulence input file found. Assuming zero microturbulence...'
     return
  endif
  !
  ! Message
  !
  write(stdo,*) 'Reading microturbulence...'
  call flush(stdo)
  !
  ! Switch formatted/unformatted
  !
  if(fex1) then
     !
     ! Open formatted
     !
     open(unit=1,file=filename1)
     !
     ! Read format number
     !
     read(1,*) iformat
     !
     ! Read number of grid points
     !
     read(1,*) nc
     !
     ! Check if this is equal to that of the grid.inp
     !
     if(nc.ne.nrcellsinp) then
        write(stdo,*) 'ERROR in line module: nr of grid points specified in '
        write(stdo,*) '    ',filename1(1:len_trim(filename1)),' unequal to the'
        write(stdo,*) '     nr specified in the grid file.'
        stop
     endif
     !
     ! Create the microturbulence arrays
     !
     allocate(lines_microturb(1:nrcells),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate the lines_microturb() array'
        stop
     endif
     !
     ! Now do the big loop
     !
     call amr_resetcount()
     call amr_nextcell(index)
     do while(index.ge.0) 
        read(1,*) dummy
        if(index.gt.0) lines_microturb(index) = dummy
        call amr_nextcell(index)
     enddo
     !
     ! Close
     !
     close(1)
  else
     !
     ! Create the microturbulence arrays
     !
     allocate(lines_microturb(1:nrcells),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate the lines_microturb() array'
        stop
     endif
     !
     ! Unformatted input for the number density of this species
     !
     if(fex2.and.fex3) then
        write(stdo,*) 'ERROR: Found both ',trim(filename2),' and ',&
             trim(filename3)
        stop
     endif
     if(fex2) then
        filename = filename2
     else
        filename = filename3
     endif
     open(unit=1,file=filename,status='old',form='unformatted')
     read(1) iiformat,reclen
     if(iiformat.ne.1) then
        write(stdo,*) 'ERROR: Format number of '//TRIM(filename)//' is invalid/unknown.'
        write(stdo,*) 'Format number = ',iiformat
        write(stdo,*) 'Record length = ',reclen
        stop
     endif
     read(1) nn
     if(nn.ne.nrcellsinp) then
        write(stdo,*) 'ERROR: '//trim(filename)//' does not have same number'
        write(stdo,*) '       of cells as the grid.'
        write(stdo,*) nn,nrcellsinp
        stop
     endif
     !
     ! Prepare for the record-based reading. NOTE: This is a nasty
     ! feature of FORTRAN 90/95: It cannot read purely binary data.
     ! It uses instead "record-based unformatted data", which is the
     ! last remaining stone-age fortran feature still in F90/95. In
     ! Fortran 2003 this is fixed: There a "stream-based unformatted"
     ! file format is possible, which is similar to the unformatted
     ! data of C or C++. In principle many F90/95 compilers already
     ! have this F03 feature implemented, but for portability we do
     ! not yet rely on it. So instead we divide the main data into
     ! "records" of a (preferably) large size, though the size cannot
     ! exceed 256^2 = 65536 bytes. The record length should be given
     ! at the start of the file: the reclen variable (which is a
     ! long integer = 8 bytes, as all other integers). The user must
     ! make sure that this is precisely the number of variables per
     ! record times the length of the float. For .uinp files (which
     ! are double precision) the length per variable is 8 bytes, and
     ! for .usinp files (which are single precision) the length per
     ! variable is 4 bytes. The "reclend" variable then gives for 
     ! how many cells there is information per record.
     !
     if(allocated(data)) deallocate(data)
     if(allocated(datas)) deallocate(datas)
     if(fex2) then
        reclend=reclen/8
        allocate(data(1:reclend),STAT=ierr)
     else
        reclend=reclen/4
        allocate(datas(1:reclend),STAT=ierr)
        allocate(data(1:reclend),STAT=ierr)
     endif
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate data array'
        stop
     endif
     call amr_resetcount()
     call amr_nextcell(index)
     do irec=1,(nrcellsinp-1)/reclend+1
        if(fex2) then
           read(1) data
        else
           read(1) datas
           data(:) = datas(:)
        endif
        i = 1
        do while((index.ge.0).and.(i.le.reclend))
           if(index.gt.0) lines_microturb(index) = data(i)
           i     = i + 1
           call amr_nextcell(index)
        enddo
     enddo
     !
     ! Close
     !
     close(1)
     if(allocated(data)) deallocate(data)
     if(allocated(datas)) deallocate(datas)
  endif
  !
  ! Determine the maximum absolute velocity. This is useful for when we
  ! have to figure out which lines may contribute to a certain wavelength.
  !
  lines_maxturbc = 0.d0
  do icell=1,nrcells
     index = cellindex(icell)
     if(lines_microturb(index).gt.lines_maxturbc) then
        lines_maxturbc = lines_microturb(index)
     endif
  enddo
  lines_maxturbc = lines_maxturbc / cc
  !
end subroutine read_microturbulence





!============================================================================
!                     THE RADIATIVE TRANSFER ROUTINES
!
!               SOME OF THESE ROUTINES CAN ALSO BE USED FOR
!                OPEN MP SHARED MEMORY PARALLELIZATION AND
!              FOR FUTURE NVIDIA GRAPHICS CARD IMPLEMENTATION
!============================================================================


!----------------------------------------------------------------------------
!                SET THE LINE TRANSFER FREQUENCY WINDOW
!----------------------------------------------------------------------------
subroutine lines_serial_init_raytrace(action)
  implicit none
  integer :: action
  ray_ns    = 1
  ray_nsmax = 1
  call lines_ray1d_init_raytrace(action)
end subroutine lines_serial_init_raytrace



!----------------------------------------------------------------------------
!                SET THE LINE TRANSFER FREQUENCY WINDOW
!----------------------------------------------------------------------------
subroutine lines_ray1d_init_raytrace(action)
  implicit none
  integer :: ierr,action
  !
  ! Action=0 means do nothing, action=1 means read if not yet read,
  ! action=2 means re-read.
  !
  if(action.eq.0) then
     return
  elseif(action.eq.1) then
     if(allocated(lines_ray_doppler)) return
  endif
  !
  ! Check lines
  !
  if(.not.rt_incl_lines) then
     write(stdo,*) 'ERROR: Trying to initialize lines, but incl_lines=F'
     stop
  endif
  !
  ! First clean up things
  !
  if(allocated(lines_ray_levpop)) deallocate(lines_ray_levpop)
  if(allocated(lines_ray_temp)) deallocate(lines_ray_temp)
  if(allocated(lines_ray_turb)) deallocate(lines_ray_turb)
  if(allocated(lines_ray_doppler)) deallocate(lines_ray_doppler)
  lines_warn_lineleap = .false.
  !
  ! Check some basic things
  !
  if(ray_nsmax.lt.1) then
     write(stdo,*) 'ERROR in line module: ray_nsmax not set.'
     stop
  endif
  !
  ! Check if the rest is allocated
  !
  if(.not.allocated(lines_nrlines_subset)) then
     write(stdo,*) 'ERROR in line module: must first initalize the line module'
     write(stdo,*) '      before setting up the line ray-tracing.'
     stop
  endif
  !
  ! Allocate the various arrays
  !  
  allocate(lines_ray_levpop(lines_maxnrlevels,lines_nr_species,ray_nsmax),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_ray_levpop(:,:,:).'
     stop 
  endif
  allocate(lines_ray_temp(ray_nsmax),STAT=ierr) 
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_ray_temp(:).'
     stop 
  endif
  allocate(lines_ray_turb(ray_nsmax),STAT=ierr) 
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_ray_turb(:).'
     stop 
  endif
  allocate(lines_ray_doppler(ray_nsmax),STAT=ierr) 
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in lines module: Could not allocate '
     write(stdo,*) '      lines_ray_doppler(:).'
     stop 
  endif
  !
end subroutine lines_ray1d_init_raytrace



!----------------------------------------------------------------------------
!        FIND ACTIVE LINES AND LEVELS FOR GIVEN WAVELENGTH GRID
!----------------------------------------------------------------------------
subroutine lines_find_active_lines_levels(nrfreq,&
                        frequencies,inu0,inu1,maxshift)
  implicit none
  integer :: nrfreq,inu0,inu1,inu,ilinesub,ispec,iline,ilevel,ilevelsub
  double precision :: frequencies(1:nrfreq),maxshift,dummy
  logical :: notdone
  !
  if(debug_check_all.ne.0) then
     if((.not.allocated(active_levels)).or.&
        (.not.allocated(active_levels_subsetindex)).or.&
        (.not.allocated(active_lines)).or.&
        (.not.allocated(active_dummy))) then
        write(stdo,*) 'ERROR in line module: active lines/levels/dummy not allocated.'
        stop
     endif
  endif
  !
  do ispec=1,lines_nr_species
     do ilevel=1,lines_maxnrlevels
        active_dummy(ilevel) = 0
     enddo
     active_nrlines(ispec) = 0
     ! BUGFIX 22.11.09: Swapped do ilinesub and do inu, and added notdone flag
     do ilinesub=1,lines_nrlines_subset(ispec)
        iline = lines_lines_subset(ilinesub,ispec)
        notdone = .true.
        do inu=inu0,inu1
           if(notdone) then
              dummy = abs(1.d0-frequencies(inu)/lines_nu0(iline,ispec))
              if(dummy.lt.maxshift) then
                 active_nrlines(ispec) = active_nrlines(ispec) + 1
                 active_lines(active_nrlines(ispec),ispec) = iline
                 ! write(stdo,*)  '   In image: including Molecule', ispec, '  Transition',iline
                 ilevel=lines_levelup(iline,ispec)
                 active_dummy(ilevel) = 1
                 ilevel=lines_leveldown(iline,ispec)
                 active_dummy(ilevel) = 1
                 notdone = .false.
              endif
           endif
        enddo
     enddo
     active_nrlevels(ispec) = 0
     do ilevelsub=1,lines_nrlevels_subset(ispec)
        ilevel = lines_levels_subset(ilevelsub,ispec)
        if(active_dummy(ilevel).eq.1) then
           active_nrlevels(ispec) = active_nrlevels(ispec) + 1
           active_levels(active_nrlevels(ispec),ispec) = ilevel
           active_levels_subsetindex(active_nrlevels(ispec),ispec) = ilevelsub
        endif
     enddo
  enddo
end subroutine lines_find_active_lines_levels


!----------------------------------------------------------------------------
!     COMPUTE THE SOURCE AND OPACITY FROM THE LEVEL POPULATIONS
!                  FOR RAY-TRACING METHOD A (SERIAL)
!
! This routine gives the full j_nu and alpha_nu for the lines. If the ray is
! only 1 element, then this can be considered as a point-evaluation. The
! values will be calculated for a predefined set of frequencies. 
! NOTE: The src and alp of the lines will be ADDED to anything that
!       already is in the src and alp (for instance dust emission!)
!----------------------------------------------------------------------------
subroutine lines_serial_addto_jnu_alpnu(nrfreq,inu0,inu1,freq,src,alp)
  implicit none
  integer :: inu,iline,ilinesub,ierr,ispec,nrfreq,inu0,inu1
  doubleprecision :: src(1:nrfreq),alp(1:nrfreq),freq(1:nrfreq)
  doubleprecision :: dummy,nup,ndown
  doubleprecision :: bdu,bud,gratio,a,dnu,phi
  doubleprecision :: dum,dum1,dum2
  doubleprecision :: const1,const2
  parameter(const1=hh/(4*pi))
  parameter(const2=cc*cc/(2*hh))
  !
  ! Checks
  !
  if(debug_check_all.ne.0) then
     if((.not.allocated(lines_ray_levpop)).or. &
          (.not.allocated(lines_ray_temp)).or. &
          (.not.allocated(lines_ray_turb)).or. &
          (.not.allocated(lines_ray_doppler))) then
        write(stdo,*) 'ERROR in line module: ray arrays not allocated.'
        stop
     endif
  endif
  !
  ! Now loop over frequency
  !
  do inu=inu0,inu1
     !
     ! Loop over actively contributing lines
     !
     do ispec=1,lines_nr_species
        do ilinesub=1,active_nrlines(ispec)
           !
           ! Get line index
           !
           iline = active_lines(ilinesub,ispec)
           !
           ! Find the upper and lower level population
           !
           nup   = lines_ray_levpop(lines_levelup_subsetindex(iline,ispec),ispec,1)
           ndown = lines_ray_levpop(lines_leveldown_subsetindex(iline,ispec),ispec,1)
           !
           ! Ratio of degenerations
           !
           gratio = lines_level_gdeg(lines_levelup(iline,ispec),ispec) /   &
                    lines_level_gdeg(lines_leveldown(iline,ispec),ispec)
           !
           ! Determine the value of the line profile function
           !
           ! Note: doppler positive means moving toward observer, hence
           !       the minus sign
           !
           dnu   = freq(inu) - lines_nu0(iline,ispec) - &
                   lines_ray_doppler(1)*lines_nu0(iline,ispec)
!           a     = lines_ray_turb(1) + sqrt(2*kk*lines_ray_temp(1)/lines_umass(ispec))
           a     = sqrt(lines_ray_turb(1)**2 + 2*kk*lines_ray_temp(1)/lines_umass(ispec))
           phi   = exp(-(cc*dnu/(lines_nu0(iline,ispec)*a))**2) * &
                   cc / (a*lines_nu0(iline,ispec)*sqrtpi)
           !
           ! The j_nu is now:
           !
           !                 h nu_0
           !          j_nu = ------ n_up A_ud phi(nu)
           !                  4 pi
           !
           src(inu) = src(inu) +                                   &
                const1 * lines_nu0(iline,ispec) * phi *            &
                lines_aud(iline,ispec) * nup 
           !
           ! Compute the Bud and Bdu
           !
           bud = lines_aud(iline,ispec) * const2 / lines_nu0(iline,ispec)**3
           bdu = bud * gratio
           !
           ! The alpha_nu is now:
           !
           !                 h nu_0
           !      alpha_nu = ------ ( n_down B_du - n_up B_ud ) phi(nu)
           !                  4 pi
           !
           alp(inu) = alp(inu) +                                   &
                const1 * lines_nu0(iline,ispec) * phi *            &
                ( bdu * ndown - bud * nup )
           !
        enddo
     enddo
     !
     ! Check that the opacity is non-negative. We do not allow masers!
     !
     if(alp(inu).lt.0.d0) then
        write(stdo,*) 'PROBLEM: Encountered negative opacity in lines!'
        write(stdo,*) '         Aborting...'
        stop 
     endif
  enddo
  !
end subroutine lines_serial_addto_jnu_alpnu


!
!    *** As long as we don't use the below subroutine, I comment it out ***
!
! !----------------------------------------------------------------------------
! !     COMPUTE THE SOURCE AND OPACITY FROM THE LEVEL POPULATIONS
! !            FOR RAY-TRACING METHOD B (1-D PREPARED RAY)
! !
! ! This routine gives the full j_nu and alpha_nu for the lines. Here it is
! ! done along a 1-D prepared ray. This is useful for future parallelization
! ! on a GPU.
! ! NOTE: The src and alp of the lines will be ADDED to anything that
! !       already is in the src and alp (for instance dust emission!)
! !----------------------------------------------------------------------------
! subroutine lines_ray1d_addto_jnu_alpnu(nrfreq,inu0,inu1,ns,freq,src,alp)
!   implicit none
!   integer :: is,inu,iline,ilinesub,ierr,ispec,nrfreq,ns,inu0,inu1
!   doubleprecision :: src(1:nrfreq,1:ns),alp(1:nrfreq,1:ns),freq(1:nrfreq)
!   doubleprecision :: dummy,nup,ndown,maxshift,tmax
!   doubleprecision :: bdu,bud,gratio,a,dnu,phi
!   doubleprecision :: dum,dum1,dum2
!   doubleprecision :: const1,const2
!   parameter(const1=hh/(4*pi))
!   parameter(const2=cc*cc/(2*hh))
!   !
!   ! Checks
!   !
!   if(debug_check_all.ne.0) then
!      if((.not.allocated(lines_ray_levpop)).or. &
!           (.not.allocated(lines_ray_temp)).or. &
!           (.not.allocated(lines_ray_turb)).or. &
!           (.not.allocated(lines_ray_doppler))) then
!         write(stdo,*) 'ERROR in line module: ray arrays not allocated.'
!         stop
!      endif
!      if(ns.ne.ray_ns) stop 5455
!   endif
!   !
!   ! Determine maximum temperature along the ray
!   !
!   tmax = 0.d0
!   do is=1,ns
!      tmax = max(tmax,lines_ray_temp(is))
!   enddo
!   !
!   ! Check and compute the maximum doppler shift + max thermal shift
!   ! The latter is multiplied by widthmargin to ensure that also the
!   ! thermal line wings are inside.
!   !
!   maxshift = 0.d0
!   do is=1,ns
!      maxshift = max(maxshift,abs(lines_ray_doppler(is)),lines_widthmargin*lines_ray_turb(is)/cc)
!   enddo
!   maxshift = maxshift + lines_widthmargin*sqrt(2*kk*tmax/lines_umass_min)/cc
!   !
!   ! Check if anywhere a line has possibly doppler-leapt farther than the
!   ! intrinsic line width.
!   !
!   do is=2,ns-1
!      dum1 = abs( lines_ray_doppler(is) - lines_ray_doppler(is-1) )
!      dum2 = abs( lines_ray_doppler(is+1) - lines_ray_doppler(is) )
!      dum  = cc * max(dum1,dum2) / &
!!!             ( sqrt(2*kk*lines_ray_temp(is)/lines_umass_max) + lines_ray_turb(is) ) 
!               sqrt(2*kk*lines_ray_temp(is)/lines_umass_max + lines_ray_turb(is)**2) 
!      if(dum.gt.lines_maxdoppler) lines_warn_lineleap=.true.
!   enddo
!   !
!   ! Check which lines are actively contributing to the current wavelength
!   !
!   call lines_find_active_lines_levels(nrfreq,freq,inu0,inu1,maxshift)
!   !
!   ! Now loop over frequency
!   !
!   ! ############## FUTURE: FOR NVIDIA GRAPHICS CARD: PARALLELIZE IN NU HERE #########
!   !
!   do inu=inu0,inu1
!      !
!      ! Now go along the ray to define the src and alp
!      !
!      do is=1,ray_ns
!         !
!         ! Loop over actively contributing lines
!         !
!         do ispec=1,lines_nr_species
!            do ilinesub=1,active_nrlines(ispec)
!               !
!               ! Get line index
!               !
!               iline = active_lines(ilinesub,ispec)
!               !
!               ! Find the upper and lower level population
!               !
!               nup   = lines_ray_levpop(lines_levelup_subsetindex(iline,ispec),ispec,is)
!               ndown = lines_ray_levpop(lines_leveldown_subsetindex(iline,ispec),ispec,is)
!               !
!               ! Ratio of degenerations
!               !
!               gratio = lines_level_gdeg(lines_levelup(iline,ispec),ispec) /   &
!                        lines_level_gdeg(lines_leveldown(iline,ispec),ispec)
!               !
!               ! Determine the value of the line profile function
!               !
!               ! Note: doppler positive means moving toward observer, hence
!               !       the minus sign
!               !
!               dnu   = freq(inu) - lines_nu0(iline,ispec) - &
!                       lines_ray_doppler(is)*lines_nu0(iline,ispec)
!!               a     = lines_ray_turb(is) + sqrt(2*kk*lines_ray_temp(is)/lines_umass(ispec))
!               a     = sqrt(lines_ray_turb(is)**2 + 2*kk*lines_ray_temp(is)/lines_umass(ispec))
!               phi   = exp(-(cc*dnu/(lines_nu0(iline,ispec)*a))**2) * &
!                       cc / (a*lines_nu0(iline,ispec)*sqrtpi)
!               !
!               ! The j_nu is now:
!               !
!               !                 h nu_0
!               !          j_nu = ------ n_up A_ud phi(nu)
!               !                  4 pi
!               !
!               src(inu,is) = src(inu,is) +                             &
!                    const1 * lines_nu0(iline,ispec) * phi *            &
!                    lines_aud(iline,ispec) * nup 
!               !
!               ! Compute the Bud and Bdu
!               !
!               bud = lines_aud(iline,ispec) * const2 / lines_nu0(iline,ispec)**3
!               bdu = bud * gratio
!               !
!               ! The alpha_nu is now:
!               !
!               !                 h nu_0
!               !      alpha_nu = ------ ( n_down B_du - n_up B_ud ) phi(nu)
!               !                  4 pi
!               !
!               alp(inu,is) = alp(inu,is) +                             &
!                    const1 * lines_nu0(iline,ispec) * phi *            &
!                    ( bdu * ndown - bud * nup )
!               !
!            enddo
!         enddo
!      enddo
!   enddo
!   !
! end subroutine lines_ray1d_addto_jnu_alpnu



!----------------------------------------------------------------------------
!     COMPUTE THE SOURCE AND OPACITY FROM THE LEVEL POPULATIONS
!               FOR THE DOPPLER-CATCHING METHOD
!
! This is the line stuff without the line profile. This is needed for the
! line "doppler catching" method. Basically everything of the line is
! calculated, except the final bit: the multiplication with the line 
! profile. The reason is that we will only know the line profile once we
! know exactly the gas velocity, and in the doppler catching method this
! is precisely the thing that is going to be varied in small sub-steps
! along a ray-element to avoid doppler jumps.
!----------------------------------------------------------------------------
subroutine lines_jnu_anu_base(nlines,jbase,alpbase)
  implicit none
  integer :: iline,ilinesub,ispec,nlines,icnt
  doubleprecision :: jbase(1:nlines),alpbase(1:nlines)
  doubleprecision :: nup,ndown
  doubleprecision :: bdu,bud,gratio
  doubleprecision :: const1,const2
  parameter(const1=hh/(4*pi))
  parameter(const2=cc*cc/(2*hh))
  !
  ! Checks
  !
  if(debug_check_all.ne.0) then
     if((.not.allocated(lines_ray_levpop)).or. &
          (.not.allocated(lines_ray_temp)).or. &
          (.not.allocated(lines_ray_turb)).or. &
          (.not.allocated(lines_ray_doppler))) then
        write(stdo,*) 'ERROR in line module: ray arrays not allocated.'
        stop
     endif
  endif
  !
  ! Loop over actively contributing lines
  !
  icnt = 0
  do ispec=1,lines_nr_species
     do ilinesub=1,active_nrlines(ispec)
        !
        ! Increase counter
        !
        icnt  = icnt + 1
        if(icnt.gt.nlines) then
           write(stdo,*) 'INTERNAL ERROR: In preparing doppler catch: Nr of lines inconsistent.'
           stop
        endif
        !
        ! Get line index
        !
        iline = active_lines(ilinesub,ispec)
        !
        ! Find the upper and lower level population
        !
        nup   = lines_ray_levpop(lines_levelup_subsetindex(iline,ispec),ispec,1)
        ndown = lines_ray_levpop(lines_leveldown_subsetindex(iline,ispec),ispec,1)
        !
        ! Ratio of degenerations
        !
        gratio = lines_level_gdeg(lines_levelup(iline,ispec),ispec) /   &
                 lines_level_gdeg(lines_leveldown(iline,ispec),ispec)
        !
        ! Compute the Bud and Bdu
        !
        bud = lines_aud(iline,ispec) * const2 / lines_nu0(iline,ispec)**3
        bdu = bud * gratio
        !
        ! The j_base is now 
        !
        !               h nu_0
        !      j_base = ------ n_up A_ud
        !                4 pi
        !
        jbase(icnt) = const1 * lines_nu0(iline,ispec) * lines_aud(iline,ispec) * nup 
        !
        ! The alpha_base is now:
        !
        !                   h nu_0
        !      alpha_base = ------ ( n_down B_du - n_up B_ud ) 
        !                    4 pi
        !
        alpbase(icnt) = const1 * lines_nu0(iline,ispec) * ( bdu * ndown - bud * nup )
        !
        !
        ! Check that the opacity is non-negative. We do not allow masers!
        !
        if(alpbase(icnt).lt.0.d0) then
           write(stdo,*) 'PROBLEM: Encountered negative opacity in lines!'
           write(stdo,*) '         Aborting...'
           stop 
        endif
     enddo
  enddo
  !
end subroutine lines_jnu_anu_base


!----------------------------------------------------------------------------
!                  MAKE ANU AND JNU BASE FROM ANU BASE
!
! This subroutine is used for the doppler catching method. The idea of the
! doppler catching method of ray-tracing is that during the ray-tracing it
! is continuously checked if a ray element doppler-jumps over a line. This
! can happen if the Delta (v.n), where n is the vector along the ray and v
! is the gas velocity, is larger than the local line width (consisting of
! thermal and microturbulent line width). To solve this, the ray element is
! then chopped into sub-elements, and the integration is done over all these
! sub-elements. The subroutine lines_jnu_anu_from_base() computes the source
! terms for all these sub-elements, starting from precalculated values such
! jbase and alpbase for each of the lines that may contribute, where alpbase
! is alpha_nu/phi(nu), i.e. the extinction coeff without the line function,
! and similar for jbase. Both can be computed already beforehand, and only
! the velocity is then needed to compute the current value of the line
! profile and thus the j_nu and alpha_nu.
! ----------------------------------------------------------------------------
subroutine lines_jnu_anu_from_base(nlines,jbase,alpbase,doppler,&
                                  turb,temp,nu,anu,jnu)
  implicit none
  integer :: icnt,nlines,iline,ispec,ilinesub
  doubleprecision :: jbase(1:nlines),alpbase(1:nlines)
  double precision :: nu,anu,jnu,doppler,turb,temp,dnu,phi,a
  icnt = 0
  anu  = 0.d0
  jnu  = 0.d0
  do ispec=1,lines_nr_species
     !
     ! Compute the line width from (unresolved=subgrid) turbulence and 
     ! the gas temperature, assuming that these are the only intrinsic 
     ! broadning processes
     !
!     a     = turb + sqrt(2*kk*temp/lines_umass(ispec))
     a     = sqrt(turb**2 + 2*kk*temp/lines_umass(ispec))
     !
     ! Only include this molecule if a.gt.0. A zero a means that the molecule
     ! is not included here.
     !
     if(a.gt.0.d0) then
        !
        ! Now for each line of this molecule
        !
        do ilinesub=1,active_nrlines(ispec)
           !
           ! Increase counter
           !
           icnt  = icnt + 1
           if(icnt.gt.nlines) then
              write(stdo,*) 'INTERNAL ERROR: In preparing doppler catch: Nr of lines inconsistent. Warn author.'
              stop
           endif
           !
           ! Get line index
           !
           iline = active_lines(ilinesub,ispec)
           !
           ! Make line profile
           !
           dnu   = nu - lines_nu0(iline,ispec) - &
                        doppler*lines_nu0(iline,ispec)
           phi   = exp(-(cc*dnu/(lines_nu0(iline,ispec)*a))**2) * &
                         cc / (a*lines_nu0(iline,ispec)*sqrtpi)
           !
           ! Add all line contributions to the anu and jnu
           !
           anu  = anu + alpbase(icnt) * phi
           jnu  = jnu + jbase(icnt) * phi
           !
        enddo
     else
        icnt = icnt + active_nrlines(ispec)
     endif
  enddo
  !
  ! Do a stupidity check
  !
  if(icnt.ne.nlines) then
     write(stdo,*) 'INTERNAL ERROR while computing a_nu and j_nu for lines from base:'
     write(stdo,*) '      nr of active lines mismatch. Warn author.'
     write(stdo,*) '      icnt = ',icnt,' nlines = ',nlines
     stop
  endif
end subroutine lines_jnu_anu_from_base


!----------------------------------------------------------------------------
!              FOR CONVENIENCE OF THE USER: WRITE LINE INFO
!----------------------------------------------------------------------------
subroutine lines_print_lineinfo(iline,ispec)
  implicit none
  integer :: iline,ispec
  call read_linedata(1)
  if(ispec.lt.1.or.ispec.gt.lines_nr_species) then
     write(stdo,*) 'ERROR: ispec out of range for lines'
     write(stdo,*) ispec
     stop
  endif
  if(iline.lt.1.or.iline.gt.lines_nrlines(ispec)) then
     write(stdo,*) 'ERROR: iline out of range for lines'
     write(stdo,*) iline
     stop
  endif
  open(unit=1,file='lineinfo.out')
  write(1,*) 1,    '            ; iformat'
  write(1,*) lines_nr_species,'            ; total nr of species'
  write(1,*) ispec,'            ; ispec'
  write(1,*) trim(lines_speciesname(ispec)),'                      ; species name'
  write(1,*) lines_umass(ispec)/mp,'  ; species mass in units of m_p'
  write(1,*) lines_maxnrlevels,'            ; max nr of levels'
  write(1,*) lines_nrlevels(ispec),'            ; total nr of levels for this species'
  write(1,*) lines_maxnrlines,'            ; max nr of lines'
  write(1,*) lines_nrlines(ispec),'            ; total nr of lines for this species'
  write(1,*) iline,'            ; iline'
  write(1,*) trim(lines_level_name(lines_levelup(iline,ispec),ispec)),'                       ; upper level name'
  write(1,*) trim(lines_level_name(lines_leveldown(iline,ispec),ispec)),'                       ; lower level name'
  write(1,*) 1d4*cc/lines_nu0(iline,ispec),'  ; line center wavelength in micron'
  close(1)
end subroutine lines_print_lineinfo

end module lines_module
